import logging
import traceback
from os import mkdir, rmdir
from pathlib import Path

import click

from pytransifex.api import Transifex
from pytransifex.config import CliSettings

client = Transifex(defer_login=True)


def path_to_slug(file_paths: list[Path]) -> list[str]:
    keep_last = (str(f).split("/")[-1] for f in file_paths)
    remove_dot = (s.split(".")[0] for s in keep_last)
    return list(remove_dot)


def extract_files(input_dir: Path) -> tuple[list[Path], list[str], str]:
    files = list(Path.iterdir(input_dir))
    slugs = path_to_slug(files)
    files_status_report = "\n".join(
        (f"file: {file} => slug: {slug}" for slug, file in zip(slugs, files))
    )
    return (files, slugs, files_status_report)


@click.group
def cli():
    pass


@click.option("-v", "--verbose", is_flag=True, default=False)
@click.option("-out", "--output-directory", is_flag=False)
@click.option("-in", "--input-directory", is_flag=False)
@click.option("-org", "--organization-slug", is_flag=False)
@click.option("-p", "--project-slug", is_flag=False)
@cli.command(
    "init", help="Initialize the CLI with the appropriate configuration values."
)
def init(**opts):
    reply = ""
    settings = CliSettings.extract_settings(**opts)
    has_to_create_dir = not Path.exists(settings.output_directory)

    try:
        click.echo(f"Initializing...")

        if has_to_create_dir:
            mkdir(settings.output_directory)

        reply += f"Initialized project with the following settings: {settings.project_slug} and saved file to {settings.config_file} "

        if not settings.input_directory:
            reply += f"WARNING: You will need to declare an input directory if you plan on using 'pytx push', as in 'pytx push --input-directory <PATH/TO/DIRECTORY>'."

    except Exception as error:
        reply += (
            f"cli:init > Failed to initialize the CLI, this error occurred: {error}."
        )

        if has_to_create_dir:
            rmdir(settings.output_directory)

        reply += f"Removed {settings.output_directory}. "

    finally:
        click.echo(reply)
        settings.to_disk()


@click.option("-in", "--input-directory", is_flag=False)
@cli.command("push", help="Push translation strings")
def push(input_directory: str | None):
    reply = ""
    settings = CliSettings.from_disk()
    input_dir = (
        Path.cwd().joinpath(input_directory)
        if input_directory
        else getattr(settings, "input_directory", None)
    )

    if not input_dir:
        raise FileExistsError(
            "cli:push > To use this 'push', you need to initialize the project with a valid path to the directory containing the files to push; alternatively, you can call this commend with 'pytx push --input-directory <PATH/TO/DIRECTORY>'."
        )

    try:
        files, slugs, files_status_report = extract_files(input_dir)
        click.echo(
            f"cli:push > Pushing {files_status_report} to Transifex under project {settings.project_slug}."
        )
        client.push(
            project_slug=settings.project_slug,
            resource_slugs=slugs,
            path_to_files=files,
        )
    except Exception as error:
        reply += f"cli:push > Failed because of this error: {error}"
        logging.error(f"traceback: {traceback.print_exc()}")
    finally:
        click.echo(reply)
        settings.to_disk()


@click.option("-l", "--only-lang", default="all")
@click.option("-out", "--output-directory", is_flag=False)
@cli.command("pull", help="Pull translation strings")
def pull(output_directory: str | Path | None, only_lang: str | None):
    reply = ""
    settings = CliSettings.from_disk()
    language_codes = only_lang.split(",") if only_lang else []

    if output_directory:
        output_directory = Path(output_directory)
        settings.output_directory = output_directory
    else:
        output_directory = settings.output_directory

    resource_slugs = []
    try:
        click.echo(
            f"Pulling translation strings ({language_codes}) from project {settings.project_slug} to {str(output_directory)}..."
        )
        client.pull(
            project_slug=settings.project_slug,
            resource_slugs=resource_slugs,
            language_codes=language_codes,
            output_dir=output_directory,
        )
    except Exception as error:
        reply += f"cli:pull > failed because of this error: {error}"
        logging.error(f"traceback: {traceback.print_exc()}")
    finally:
        click.echo(reply)
        settings.to_disk()



from abc import ABC, abstractmethod
from typing import Any


class Tx(ABC):
    # TODO
    # This interface modelled after api.py:Transifex satisfies the expectations of qgis-plugin-cli
    # but it falls short of requiring an implementation for methods that qgis-plugin-cli does not use:
    # { coordinator, create_translation, delete_project, delete_resource, delete_team }

    @abstractmethod
    def create_project(
        self,
        project_slug: str,
        project_name: str | None = None,
        source_language_code: str = "en-gb",
        outsource_project_name: str | None = None,
        private: bool = False,
        repository_url: str | None = None,
    ):
        raise NotImplementedError

    def delete_project(self, project_slug: str):
        ...

    @abstractmethod
    def list_resources(self, project_slug: str) -> list[Any]:
        raise NotImplementedError

    def delete_team(self, team_slug: str):
        ...

    @abstractmethod
    def create_resource(
        self,
        project_slug: str,
        path_to_file: str,
        resource_slug: str | None = None,
        resource_name: str | None = None,
    ):
        raise NotImplementedError

    def delete_resource(self, project_slug: str, resource_slug: str):
        ...

    @abstractmethod
    def update_source_translation(
        self, project_slug: str, resource_slug: str, path_to_file: str
    ):
        raise NotImplementedError

    def create_translation(
        self, project_slug: str, language_code: str, path_to_file: str
    ) -> dict[str, Any]:
        ...

    @abstractmethod
    def get_translation(
        self, project_slug: str, resource_slug: str, language: str, path_to_file: str
    ):
        raise NotImplementedError

    @abstractmethod
    def list_languages(self, project_slug: str, resource_slug: str) -> list[Any]:
        raise NotImplementedError

    @abstractmethod
    def create_language(self, project_slug: str, path_to_file: str, resource_slug: str):
        raise NotImplementedError

    def coordinator(self, project_slug: str, language_code="str") -> str:
        ...

    @abstractmethod
    def project_exists(self, project_slug: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def ping(self) -> bool:
        raise NotImplementedError



import unittest
from functools import partial
from time import sleep as tsleep

from pytransifex.utils import concurrently


def fn(a: int, b: int) -> int:
    tsleep(a)
    return a + b


class TestUtils(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        it = [1, 2, 3]
        cls.it = it
        cls.args = [tuple([a, b]) for a, b in zip(it, it)]
        cls.partials = [partial(fn, a, b) for a, b in zip(it, it)]
        cls.res = [2, 4, 6]

    def test1_map_async(self):
        res = concurrently(partials=self.partials)
        assert res == self.res

    def test2_map_async(self):
        res = concurrently(fn=fn, args=self.args)
        assert res == self.res



import unittest
from pathlib import Path
from shutil import rmtree

from pytransifex.api import Transifex
from pytransifex.interfaces import Tx


class TestNewApi(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tx = Transifex(defer_login=True)
        cls.project_slug = "test_project_pytransifex"
        cls.project_name = "Test Project PyTransifex"
        cls.resource_slug = "test_resource_fr"
        cls.resource_name = "Test Resource FR"
        cls.path_to_file = Path.cwd().joinpath("tests", "input", "test_resource_fr.po")
        cls.output_dir = Path.cwd().joinpath("tests", "output")

        if missing := next(filter(lambda p: not p.exists(), [cls.path_to_file]), None):
            raise ValueError(
                f"Unable to complete test with broken tests inputs. Found missing: {missing}"
            )

        if project := cls.tx.get_project(project_slug=cls.project_slug):
            print("Found old project, removing.")
            project.delete()

        print("Creating a brand new project")
        cls.tx.create_project(
            project_name=cls.project_name, project_slug=cls.project_slug, private=True
        )

    @classmethod
    def tearDownClass(cls):
        if Path.exists(cls.output_dir):
            rmtree(cls.output_dir)

    def test1_new_api_satisfies_abc(self):
        assert isinstance(self.tx, Tx)

    def test2_create_project(self):
        # Done in setUpClass
        pass

    def test3_list_resources(self):
        _ = self.tx.list_resources(project_slug=self.project_slug)
        assert True

    def test4_create_resource(self):
        self.tx.create_resource(
            project_slug=self.project_slug,
            resource_name=self.resource_name,
            resource_slug=self.resource_slug,
            path_to_file=self.path_to_file,
        )
        assert True

    def test5_update_source_translation(self):
        self.tx.update_source_translation(
            project_slug=self.project_slug,
            path_to_file=self.path_to_file,
            resource_slug=self.resource_slug,
        )
        assert True

    def test6_create_language(self):
        self.tx.create_language(self.project_slug, "fr_CH")

    def test7_list_languages(self):
        languages = self.tx.list_languages(project_slug=self.project_slug)
        assert languages is not None

    def test8_get_translation(self):
        self.tx.get_translation(
            project_slug=self.project_slug,
            resource_slug=self.resource_slug,
            language_code="fr_CH",
            output_dir=self.output_dir,
        )
        assert Path.exists(Path.joinpath(self.output_dir, self.resource_slug))

    def test9_project_exists(self):
        verdict = self.tx.project_exists(project_slug=self.project_slug)
        assert verdict is not None

    def test10_ping(self):
        self.tx.ping()
        assert True

    def test11_stats(self):
        stats = self.tx.get_project_stats(project_slug=self.project_slug)
        print(str(stats))
        assert stats

    def test12_stats(self):
        self.tx.get_project_stats(project_slug=self.project_slug)

    def test13_stats(self):
        self.tx.get_project_stats(project_slug=self.project_slug)


if __name__ == "__main__":
    unittest.main()

"""
# Don't remove this!
curl -g \
    --request GET --url "https://rest.api.transifex.com/resource_language_stats?filter[project]=o%3Aopengisch%3Ap%3Aqfield-documentation" \
    --header 'accept: application/vnd.api+json' \
    --header 'authorization: Bearer TOKEN'

curl -g \
    --request GET \
    --url "https://rest.api.transifex.com/resource_language_stats?filter[project]=o%3Atest_pytransifex%3Ap%3Atest_project_pytransifex" \
    --header 'accept: application/vnd.api+json' \
    --header 'authorization: Bearer TOKEN'
"""



from dataclasses import asdict, dataclass
from os import environ
from pathlib import Path
from typing import Any, NamedTuple

import toml
from dotenv import load_dotenv


class ApiConfig(NamedTuple):
    api_token: str
    organization_name: str
    i18n_type: str
    host_name = "https://rest.api.transifex.com"
    project_slug: str | None = None

    @classmethod
    def from_env(cls) -> "ApiConfig":
        load_dotenv()

        token = environ.get("TX_TOKEN")
        organization = environ.get("ORGANIZATION")
        i18n_type = environ.get("I18N_TYPE", "PO")

        if any(not v for v in [token, organization, i18n_type]):
            raise Exception(
                "Envars 'TX_TOKEN' and 'ORGANIZATION' must be set to non-empty values. Aborting now."
            )

        return cls(token, organization, i18n_type)  # type: ignore


path_keys = ["input_directory", "output_directory", "config_file"]
mandatory = ["organization_slug", "project_slug"]
defaults = {
    "output_directory": Path.cwd().joinpath("output"),
    "config_file": Path.cwd().joinpath(".pytx_config.yml"),
}


@dataclass
class CliSettings:
    organization_slug: str
    project_slug: str
    input_directory: Path | None
    output_directory: Path = defaults["output_directory"]
    config_file: Path = defaults["config_file"]

    @classmethod
    def extract_settings(cls, **user_data) -> "CliSettings":
        if missing := [k for k in mandatory if not k in user_data]:
            raise Exception(
                f"These keys are not set or do not have a well-defined value: {', '.join(missing)}"
            )
        if empty := [k for k, v in user_data.items() if k in mandatory and not v]:
            raise Exception(f"These keys have an empty value: {', '.join(empty)}")

        organization_slug = user_data["organization_slug"]
        project_slug = user_data["project_slug"]
        input_directory = user_data.get("input_directory")
        config_file = CliSettings.get_or_default("config_file", user_data)
        output_directory = CliSettings.get_or_default("output_directory", user_data)

        return cls(
            organization_slug,
            project_slug,
            input_directory,
            output_directory,
            config_file,
        )

    @classmethod
    def from_disk(cls) -> "CliSettings":
        d = toml.load(cls.config_file)
        return cls.extract_settings(**d)

    def to_disk(self):
        with open(self.config_file, "w") as fh:
            toml.dump(self.serialize(), fh)

    def serialize(self) -> dict[str, Any]:
        return {
            k: (str(v) if k in path_keys else v)
            for k, v in asdict(self).items()
            if v is not None
        }

    @staticmethod
    def deserialize(d: dict[str, Any]) -> dict[str, Any]:
        return {k: (Path(v) if k in path_keys else v) for k, v in d.items()}

    @staticmethod
    def get_or_default(k: str, obj: dict[str, Any]) -> Any:
        truthy_obj = {k: v for k, v in obj.items() if v}
        if k in truthy_obj:
            return truthy_obj[k]
        return defaults[k]



class TransifexException(Exception):
    pass


class PyTransifexException(TransifexException):
    def __init__(self, response=None):
        super(PyTransifexException, self).__init__(response)
        self.response = response

    def __str__(self):
        if self.response is None:
            return super(PyTransifexException, self).__str__()
        return "{code} from {url}: {content}".format(
            code=self.response.status_code,
            url=self.response.url,
            content=self.response.content,
        )


class InvalidSlugException(TransifexException):
    pass



import unittest
from os import remove
from pathlib import Path

from click.testing import CliRunner

from pytransifex.api import Transifex
from pytransifex.cli import cli


class TestCli(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tx = Transifex(defer_login=True)
        cls.project_slug = "test_project_pytransifex"
        cls.project_name = "Test Project PyTransifex"
        cls.resource_slug = "test_resource_fr"
        cls.resource_name = "Test Resource FR"
        cls.path_to_input_dir = Path.cwd().joinpath("tests", "input")
        cls.path_to_file = cls.path_to_input_dir.joinpath("test_resource_fr.po")
        cls.output_dir = Path.cwd().joinpath("tests", "output")

        if missing := next(
            filter(lambda p: not p.exists(), [cls.path_to_file, cls.path_to_input_dir]),
            None,
        ):
            raise ValueError(
                f"Unable to complete test with broken tests inputs. Found missing: {missing}"
            )

        if project := cls.tx.get_project(project_slug=cls.project_slug):
            print("Found old project, removing.")
            project.delete()

        print("Creating a brand new project")
        cls.tx.create_project(
            project_name=cls.project_name, project_slug=cls.project_slug, private=True
        )

        cls.runner = CliRunner()

    @classmethod
    def tearDownClass(cls):
        if Path.exists(cls.output_dir):
            remove(cls.output_dir)

    def test1_init(self):
        result = self.runner.invoke(
            cli,
            ["init", "-p", self.project_slug, "-org", "test_pytransifex"],
        )
        passed = result.exit_code == 0
        assert passed

    def test2_push(self):
        result = self.runner.invoke(cli, ["push", "-in", str(self.path_to_input_dir)])
        passed = result.exit_code == 0
        print(result.output)
        assert passed

    def test3_pull(self):
        result = self.runner.invoke(cli, ["pull", "-l", "fr_CH,en_GB"])
        passed = result.exit_code == 0
        print(result.output)
        assert passed


if __name__ == "__main__":
    unittest.main()



from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import wraps
from typing import Any, Callable


def ensure_login(f):
    @wraps(f)
    def capture_args(instance, *args, **kwargs):
        if not instance.logged_in:
            instance.login()
        return f(instance, *args, **kwargs)

    return capture_args


def concurrently(
    *,
    fn: Callable | None = None,
    args: list[Any] | None = None,
    partials: list[Any] | None = None,
) -> list[Any]:
    with ThreadPoolExecutor() as pool:
        if not partials is None:
            assert args is None and fn is None
            futures = [pool.submit(p) for p in partials]
        elif (not args is None) and (not fn is None):
            assert partials is None
            futures = [pool.submit(fn, *a) for a in args]
        else:
            raise ValueError(
                "Exactly 1 of 'partials' or 'args' must be defined. Found neither was when calling concurrently."
            )
        return [f.result() for f in as_completed(futures)]



import logging
from os import mkdir
from pathlib import Path
from typing import Any

import requests
from transifex.api import transifex_api as tx_api
from transifex.api.jsonapi import JsonApiException
from transifex.api.jsonapi.exceptions import DoesNotExist
from transifex.api.jsonapi.resources import Resource

from pytransifex.config import ApiConfig
from pytransifex.interfaces import Tx
from pytransifex.utils import concurrently, ensure_login


class Client(Tx):
    """
    The proper Transifex client expected by the cli and other consumers.
    By default instances are created and logged in 'lazyly' -- when creation or login cannot be deferred any longer.
    """

    def __init__(self, config: ApiConfig, defer_login: bool = False):
        """Extract config values, consumes API token against SDK client"""
        self.api_token = config.api_token
        self.host = config.host_name
        self.organization_name = config.organization_name
        self.i18n_type = config.i18n_type
        self.logged_in = False

        if not defer_login:
            self.login()

    def login(self):
        if self.logged_in:
            return

        # Authentication
        tx_api.setup(host=self.host, auth=self.api_token)
        self.logged_in = True

        # Saving organization and projects to avoid round-trips
        organization = tx_api.Organization.get(slug=self.organization_name)
        self.projects = organization.fetch("projects")
        self.organization = organization
        logging.info(f"Logged in as organization: {self.organization_name}")

    @ensure_login
    def create_project(
        self,
        project_slug: str,
        project_name: str | None = None,
        source_language_code: str = "en_GB",
        private: bool = False,
        *args,  # absorbing extra args
        **kwargs,  # absorbing extra kwargs
    ) -> None | Resource:
        """Create a project."""
        source_language = tx_api.Language.get(code=source_language_code)

        try:
            res = tx_api.Project.create(
                name=project_name,
                slug=project_slug,
                source_language=source_language,
                private=private,
                organization=self.organization,
            )
            logging.info("Project created!")
            return res
        except JsonApiException as error:
            if "already exists" in error.detail:  # type: ignore
                return self.get_project(project_slug=project_slug)

    @ensure_login
    def get_project(self, project_slug: str) -> None | Resource:
        """Fetches the project matching the given slug"""
        if self.projects:
            try:
                res = self.projects.get(slug=project_slug)
                logging.info("Got the project!")
                return res
            except DoesNotExist:
                return None

    @ensure_login
    def list_resources(self, project_slug: str) -> list[Any]:
        """List all resources for the project passed as argument"""
        if project := self.get_project(project_slug=project_slug):
            if resources := project.fetch("resources"):
                return list(resources.all())
            else:
                return []

        raise Exception(
            f"Unable to find any project under this organization: '{self.organization}'"
        )

    @ensure_login
    def create_resource(
        self,
        project_slug: str,
        path_to_file: Path,
        resource_slug: str | None = None,
        resource_name: str | None = None,
    ):
        """Create a resource using the given file contents, slugs and names"""
        if not (resource_slug or resource_name):
            raise Exception("Please give either a resource_slug or a resource_name")

        if project := self.get_project(project_slug=project_slug):
            resource = tx_api.Resource.create(
                project=project,
                name=resource_name or resource_slug,
                slug=resource_slug or resource_name,
                i18n_format=tx_api.I18nFormat(id=self.i18n_type),
            )

            with open(path_to_file, "r") as fh:
                content = fh.read()

            tx_api.ResourceStringsAsyncUpload.upload(content, resource=resource)
            logging.info(f"Resource created: {resource_slug or resource_name}")

        else:
            raise Exception(
                f"Not project could be found wiht the slug '{project_slug}'. Please create a project first."
            )

    @ensure_login
    def update_source_translation(
        self, project_slug: str, resource_slug: str, path_to_file: Path
    ):
        """
        Update the translation strings for the given resource using the content of the file
        passsed as argument
        """
        if not "slug" in self.organization.attributes:
            raise Exception(
                "Unable to fetch resource for this organization; define an 'organization slug' first."
            )

        logging.info(
            f"Updating source translation for resource {resource_slug} from file {path_to_file} (project: {project_slug})."
        )

        if project := self.get_project(project_slug=project_slug):
            if resources := project.fetch("resources"):
                if resource := resources.get(slug=resource_slug):
                    with open(path_to_file, "r") as fh:
                        content = fh.read()

                    tx_api.ResourceStringsAsyncUpload.upload(content, resource=resource)
                    logging.info(f"Source updated for resource: {resource_slug}")
                    return

        raise Exception(
            f"Unable to find resource '{resource_slug}' in project '{project_slug}'"
        )

    @ensure_login
    def get_translation(
        self,
        project_slug: str,
        resource_slug: str,
        language_code: str,
        output_dir: Path,
    ):
        """Fetch the translation resource matching the given language"""
        language = tx_api.Language.get(code=language_code)
        file_name = Path.joinpath(output_dir, resource_slug)

        if project := self.get_project(project_slug=project_slug):
            if resources := project.fetch("resources"):
                if resource := resources.get(slug=resource_slug):
                    url = tx_api.ResourceTranslationsAsyncDownload.download(
                        resource=resource, language=language
                    )
                    translated_content = requests.get(url).text

                    if not Path.exists(output_dir):
                        mkdir(output_dir)

                    with open(file_name, "w") as fh:
                        fh.write(translated_content)

                    logging.info(
                        f"Translations downloaded and written to file (resource: {resource_slug})"
                    )

                else:
                    raise Exception(
                        f"Unable to find any resource with this slug: '{resource_slug}'"
                    )
            else:
                raise Exception(
                    f"Unable to find any resource for this project: '{project_slug}'"
                )
        else:
            raise Exception(
                f"Couldn't find any project with this slug: '{project_slug}'"
            )

    @ensure_login
    def list_languages(self, project_slug: str) -> list[Any]:
        """
        List all languages for which there is at least 1 resource registered
        under the parameterised project
        """
        if self.projects:
            if project := self.projects.get(slug=project_slug):
                languages = project.fetch("languages")
                logging.info(f"Obtained these languages")
                return languages
            raise Exception(
                f"Unable to find any project with this slug: '{project_slug}'"
            )
        raise Exception(
            f"Unable to find any project under this organization: '{self.organization}'"
        )

    @ensure_login
    def create_language(
        self,
        project_slug: str,
        language_code: str,
        coordinators: None | list[Any] = None,
    ):
        """Create a new language resource in the remote Transifex repository"""
        if project := self.get_project(project_slug=project_slug):
            project.add("languages", [tx_api.Language.get(code=language_code)])

            if coordinators:
                project.add("coordinators", coordinators)

            logging.info(f"Created language resource for {language_code}")

    @ensure_login
    def project_exists(self, project_slug: str) -> bool:
        """Check if the project exists in the remote Transifex repository"""
        if self.projects and self.projects.get(slug=project_slug):
            return True
        return False

    @ensure_login
    def ping(self) -> bool:
        """
        Exposing this just for the sake of satisfying qgis-plugin-cli's expectations
        There is no need to ping the server on the current implementation, as connection is handled by the SDK
        """
        logging.info("'ping' is deprecated!")
        return True

    @ensure_login
    def get_project_stats(self, project_slug: str) -> dict[str, Any]:
        if self.projects:
            if project := self.projects.get(slug=project_slug):
                if resource_stats := tx_api.ResourceLanguageStats(project=project):
                    return resource_stats.to_dict()

        raise Exception(f"Unable to find translation for this project {project_slug}")

    @ensure_login
    def pull(
        self,
        project_slug: str,
        resource_slugs: list[str],
        language_codes: list[str],
        output_dir: Path,
    ):
        """Pull resources from project."""
        args = []
        for l_code in language_codes:
            for slug in resource_slugs:
                args.append(tuple([project_slug, slug, l_code, output_dir]))

        res = concurrently(
            fn=self.get_translation,
            args=args,
        )

        logging.info(f"Pulled {args} for {len(res)} results).")

    @ensure_login
    def push(
        self, project_slug: str, resource_slugs: list[str], path_to_files: list[Path]
    ):
        """Push resources with files under project."""
        if len(resource_slugs) != len(path_to_files):
            raise ValueError(
                f"Resources slugs ({len(resource_slugs)}) and Path to files ({len(path_to_files)}) must be equal in size!"
            )

        resource_zipped_with_path = list(zip(resource_slugs, path_to_files))
        resources = self.list_resources(project_slug)
        logging.info(
            f"Found {len(resources)} resource(s) for {project_slug}. Checking for missing resources and creating where necessary."
        )
        created_when_missing_resource = []

        for slug, path in resource_zipped_with_path:
            logging.info(f"Slug: {slug}. Resources: {resources}.")
            if not slug in resources:
                logging.info(
                    f"{project_slug} is missing {slug}. Creating it from {path}."
                )
                self.create_resource(
                    project_slug=project_slug, path_to_file=path, resource_slug=slug
                )
                created_when_missing_resource.append(slug)

        args = [
            tuple([project_slug, slug, path])
            for slug, path in resource_zipped_with_path
            if not slug in created_when_missing_resource
        ]

        res = concurrently(
            fn=self.update_source_translation,
            args=args,
        )

        logging.info(f"Pushed {args} for {len(res)} results.")


class Transifex:
    """
    Singleton factory to ensure the client is initialized at most once.
    Simpler to manage than a solution relying on 'imports being imported once in Python.
    """

    client = None

    def __new__(cls, *, defer_login: bool = False, **kwargs):
        if not cls.client:
            try:
                if kwargs:
                    config = ApiConfig(**kwargs)
                else:
                    logging.info(
                        f"As you called 'Transifex' without argument, we'll try defining your project from environment variables."
                    )
                    config = ApiConfig.from_env()

                cls.client = Client(config, defer_login)

            except Exception as error:
                available = list(ApiConfig._fields)
                msg = f"Unable to define a proper config. API initialization uses the following fields, with only 'project_slug' optional: {available}"
                logging.error(f"{msg}:\n{error}")

        return cls.client
